module Views.BarPlot exposing (view, init, update)

import Config.BarPlotConfig as Local exposing (Model)

import Html exposing (Html, div, input, label, select, option)
import Html.Attributes exposing (type_, style)
import Html.Events exposing (onClick, onInput)

import Model exposing (Global, View(..))
import Msg exposing (Msg(..))
import Dict exposing (..)

import Axis
import Scale exposing (BandScale, ContinuousScale, defaultBandConfig)
import TypedSvg exposing (g, rect, style, svg, text_ )
import TypedSvg.Attributes exposing (class, textAnchor, transform, viewBox)
import TypedSvg.Attributes.InPx exposing (height, width, x, y, fontSize)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Types exposing (AnchorAlignment(..), Transform(..))

import Utils exposing (codeToName)


-- Constants for width and height
w : Float
w =
    900

h : Float
h =
    600

padding : Float
padding =
    85



init : Model
init = { ubound = Nothing, multiplier = 1, visatype = "totalVisasApplied" }

propagate : Local.Msg -> Msg
propagate msg =
    GotBarPlot msg

update : Local.Msg -> Global -> Model -> ( Global, Model, Cmd Msg )
update msg global model = 
    case msg of
        Local.UboundChange ubound ->
            ( global, { model | ubound = ubound }, Cmd.none )
        Local.MultiplierChange multiplier ->
            ( global, { model | multiplier = multiplier }, Cmd.none )
        Local.VisatypeChange visatype ->
            ( global, { model | visatype = visatype }, Cmd.none )



view : Global -> Model -> Html Msg
view global model = 
    let
        -- Iterate global.data and sum totalVisasApplied/totalVisasIssued for each state
        totalList : List (String, Float)
        totalList =
            global.data
                |> List.map (\entry -> (entry.state, Maybe.withDefault 0 (if model.visatype == "totalVisasApplied" then entry.totalVisasApplied else entry.totalVisasIssued)))
                |> List.foldl
                    (\(state, count) acc ->
                        Dict.update state
                            (\maybeCount ->
                                Just (Maybe.withDefault 0 maybeCount + count)
                            )
                            acc
                    )
                    Dict.empty
                |> Dict.toList
                |> List.map (\(state, count) -> (state, count))

        -- Filter list to only include states with totalVisasApplied/totalVisasIssued >= given ubound
        filteredList : List (String, Float)
        filteredList =
            let
                fn : (String, Float) -> Bool
                fn (state, count) =
                    case model.ubound of
                        Just ubound ->
                            count < toFloat (ubound * model.multiplier)
                        Nothing ->
                            True
            in
            List.filter fn totalList

        -- PLOTTING FUNCTIONS --
        
        -- Define the scale for the x-axis (countries)
        xScale : List ( String, Float ) -> BandScale String
        xScale dat =
            List.map Tuple.first dat
                |> Scale.band { defaultBandConfig | paddingInner = 0.1, paddingOuter = 0.2 } ( 0, w - 2 * padding )

        -- Dynamically adjust the scale for the y-axis based on the max value
        yScale : List ( String, Float ) -> ContinuousScale Float
        yScale dat =
            let
                maxValue = 
                    List.maximum (List.map Tuple.second dat) 
                    |> Maybe.withDefault 0
            in
            Scale.linear ( h - 2 * padding, 0 ) ( 0, maxValue )

        -- Function to convert large numbers to K or M postfix
        largeNmbrConvert : Float -> String
        largeNmbrConvert value =
            if value >= 1000 && value < 1000000 then
                String.fromInt (round (value / 1000)) ++ "K"
            else if value >= 1000000 then
                String.fromInt (round (value / 1000000)) ++ "M"
            else
                String.fromFloat value

        -- Define the column (bar) for each country
        column : BandScale String -> ( String, Float ) -> Svg msg
        column scale ( country, value ) =
            g [ class [ "column" ] ]
                [ rect
                    [ x <| Scale.convert scale country
                    , y <| Scale.convert (yScale filteredList) value
                    , width <| Scale.bandwidth scale
                    , height <| h - Scale.convert (yScale filteredList) value - 2 * padding
                    ]
                    []
                , text_
                    [ x <| Scale.convert scale country + Scale.bandwidth scale / 2
                    , y <| Scale.convert (yScale filteredList) value - 5
                    , textAnchor AnchorMiddle
                    ]
                    [ Html.text <| (largeNmbrConvert value) ]
                ]

        -- Define the x-axis
        xAxis : List ( String, Float ) -> Svg msg
        xAxis dat =
            Axis.bottom [ Axis.tickFormat (\x -> codeToName x) ] (Scale.toRenderable identity (xScale dat))

        -- Define the y-axis
        yAxis : Svg msg
        yAxis =
            Axis.left [ Axis.tickCount 10, Axis.tickFormat largeNmbrConvert ] (yScale filteredList)


        -- Handle upper bound change
        handleUboundChange : String -> Msg
        handleUboundChange msg =
            String.toInt msg
            |> Local.UboundChange
            |> propagate

        -- Handle multiplier change
        handleMultiplierChange : String -> Msg
        handleMultiplierChange msg =
            String.toInt msg
            |> Maybe.withDefault model.multiplier
            |> Local.MultiplierChange
            |> propagate

        -- Handle visatype change
        handleVisatypeChange : String -> Msg
        handleVisatypeChange msg =
            Local.VisatypeChange msg
            |> propagate

        -- View function to render the chart
        viewPlot : List ( String, Float ) -> Svg msg
        viewPlot stateData =
            svg [ viewBox 0 0 w h ]
                [ TypedSvg.style [] [ Html.text """
                    -- .column rect { fill: rgba(2, 48, 32, 0.7); }
                    .column rect { fill: rgb(78, 154, 6); }
                    .column text { display: none; font-size: 14px }
                    -- .column:hover rect { fill: rgb(2, 48, 32); }
                    .column:hover rect { fill: rgb(57, 110, 6); }
                    .column:hover text { display: inline; }
                    .xAxis .tick text { transform: translate(-25px, 20px) rotate(-45deg);
                """ ]
                , g [ class [ "xAxis" ], transform [ Translate (padding - 1) (h - padding) ] ]
                    [ xAxis stateData ]
                , g [ class [ "yAxis" ], transform [ Translate (padding - 1) padding ] ]
                    [ yAxis ]
                , g [ transform [ Translate padding padding ], class [ "series" ] ]
                    <| List.map (column (xScale stateData)) stateData
                , text_
                    [ textAnchor AnchorMiddle
                    , x <| w / 2
                    , y <| h - padding + 80
                    , fontSize 12 ]
                    [ Html.text "Schengen State" ]
                , text_
                    [ textAnchor AnchorMiddle
                    , x <| padding
                    , y <| padding - 10
                    , fontSize 12 ]
                    [ Html.text (if model.visatype == "totalVisasApplied" then "Total Visas Applied" else "Total Visas Issued") ]
                ]
        
    in
    div [  Html.Attributes.class "barplot-wrapper"  ]
        [ div [ Html.Attributes.style "display" "flex", Html.Attributes.style "flex-direction" "column", Html.Attributes.style "gap" "6px" ] [

            div [ Html.Attributes.style "display" "flex", Html.Attributes.style "flex-direction" "row", Html.Attributes.style "gap" "6px" ]
                [ label [Html.Attributes.for "select_visatype"] [ Html.text "Choose Visa Type: " ]
                , select
                    [ Html.Attributes.id "select_visatype"
                    , Html.Attributes.style "width" "10rem"
                    , Html.Attributes.name "visatype"
                    , onInput handleVisatypeChange
                    ]
                    [ option
                        [ Html.Attributes.value "totalVisasApplied" ]
                        [ Html.text "Total Visas Applied" ]
                    , option
                        [ Html.Attributes.value "totalVisasIssued" ]
                        [ Html.text "Total Visas Issued" ]
                    ]
                ]

            , div[] [
                Html.text "Upper bound: "
                , input
                    [ Html.Attributes.type_ "number"
                    , Html.Attributes.placeholder "Enter..."
                    , Html.Attributes.min "0"
                    , Html.Attributes.value (model.ubound |> Maybe.map String.fromInt |> Maybe.withDefault "" )
                    , onInput handleUboundChange
                    ] []
                , input
                    [ Html.Attributes.id "rad_1"
                    , Html.Attributes.name "multiplier"
                    , Html.Attributes.type_ "radio"
                    , Html.Attributes.value "1"
                    , Html.Attributes.checked (model.multiplier == 1)
                    , onInput handleMultiplierChange
                    ] []
                , label [Html.Attributes.for "rad_k"] [ Html.text "×1" ]
                , input
                    [ Html.Attributes.id "rad_k"
                    , Html.Attributes.name "multiplier"
                    , Html.Attributes.type_ "radio"
                    , Html.Attributes.value "1000"
                    , Html.Attributes.checked (model.multiplier == 1000)
                    , onInput handleMultiplierChange
                    ] []
                , label [Html.Attributes.for "rad_k"] [ Html.text "×1K" ]
                , input
                    [ Html.Attributes.id "rad_m"
                    , Html.Attributes.name "multiplier"
                    , Html.Attributes.type_ "radio"
                    , Html.Attributes.value "1000000"
                    , Html.Attributes.checked (model.multiplier == 1000000)
                    , onInput handleMultiplierChange
                    ] []
                , label [Html.Attributes.for "rad_m"] [ Html.text "×1M" ]
            ]
        ]
        , viewPlot filteredList ]
